package Slim::Control::Command;

# $Id: Command.pm 2143 2005-02-20 17:40:10Z dsully $
#
# SlimServer Copyright (C) 2001-2004 Sean Adams, Slim Devices Inc.
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License,
# version 2.

use strict;

use File::Basename;
use File::Spec::Functions qw(:ALL);
use FileHandle;
use IO::Socket qw(:DEFAULT :crlf);
use Time::HiRes;
use Slim::Display::Display;
use Slim::Utils::Misc;
use Slim::Utils::Scan;
use Slim::Utils::Strings qw(string);

my %executeCallbacks;

#############################################################################
# execute - does all the hard work.  Use it.
# takes:
#   a client reference
#   a reference to an array of parameters
#   a reference to a callback function
#   a list of callback function args
#
# returns an array containing the given parameters

sub execute {
	my($client, $parrayref, $callbackf, $callbackargs) = @_;
	
	my $p0 = $$parrayref[0];
	my $p1 = $$parrayref[1];
	my $p2 = $$parrayref[2];
	my $p3 = $$parrayref[3];
	my $p4 = $$parrayref[4];
	my $p5 = $$parrayref[5];
	my $p6 = $$parrayref[6];
	my $p7 = $$parrayref[7];

	my $callcallback = 1;
	my @returnArray = ();
	my $pushParams = 1;	


	$::d_command && msg(" Executing command " . ($client ? $client->id() : "no client") . ": $p0 (" .
			(defined $p1 ? $p1 : "") . ") (" .
			(defined $p2 ? $p2 : "") . ") (" .
			(defined $p3 ? $p3 : "") . ") (" .
			(defined $p4 ? $p4 : "") . ") (" .
			(defined $p5 ? $p5 : "") . ") (" .
			(defined $p6 ? $p6 : "") . ") (" .
			(defined $p7 ? $p7 : "") . ")\n");
			
	# The first parameter is the client identifier to execute the command. Column C in the
	# table below indicates if a client is required (Y) or not (N) for the command.
	# If a parameter is "?" it is replaced by the current value in the array
	# returned by the function
	
# COMMAND LIST #

	# C	P0				P1							P2							P3			P4			P5			P6
	
# GENERAL
	# N	debug			<debugflag>					<0|1|?|>
	# N	listen			<0|1|>
	# N	pref			<prefname>					<prefvalue|?>
	# N	exit
	
# PLAYERS
	# N	player			count						?
	# N	player			<id|address>				<playerindex|playerid>		? (address form deprecated)
	# N	player			ip							<playerindex|playerid>		?
	# N	player			model						<playerindex|playerid>		?
	# N	player			name						<playerindex|playerid>		?
	# N	player            displaytype     			<playerindex|playerid>		?
	# Y	sleep 			<0..n|?>
	# Y	sync			<playerindex|playerid|-|?>
	# Y	power 			<0|1|?|>
	# Y	signalstrength 	?
	# Y	connected		?
	# Y	mixer			volume						<0..100|-100..+100|?>
	# Y	mixer			balance	(not implemented)
	# Y	mixer			bass						<0..100|-100..+100|?>
	# Y	mixer			treble						<0..100|-100..+100|?>
	# Y	mixer			pitch						<80..120|-100..+100|?>
	# Y	mixer			muting
	# Y	display			<line1>						<line2>						<duration>
	# Y	display			?							?
	# Y	displaynow		?							?
	# Y	playerpref		<prefname>					<prefvalue|?>
	# Y	button			<buttoncode>
	# Y	ir				<ircode>					<time>
	# Y	rate			<rate|?>
	
#DATABASE	
	# N	rescan	
	# N	wipecache
	# N	info			total						genres						?
	# N	info			total						artists						?
	# N	info			total						albums						?
	# N	info			total						songs						?

#PLAYLISTS
	# Y	mode			<play|pause|stop|?>	
	# Y	play		
	# Y	pause			<0|1|>	
	# Y	stop
	# Y	time|gototime	<0..n|-n|+n|?>
	# Y	genre			?
	# Y	artist			?
	# Y	album			?
	# Y	title			?
	# Y	duration		?
	# Y	path			?
		
	# Y	playlist		play|load 					<item>						[<title>] (item can be a song, playlist or directory)
	# Y	playlist		add|append					<item> (item can be a song, playlist or directory)
	# Y	playlist		insert|insertlist			<item> (item can be a song, playlist or directory)
	# Y	playlist		deleteitem					<item> (item can be a song, playlist or directory)
	# Y	playlist 		move 						<fromindex> 				<toindex>	
	# Y	playlist		delete 						<index>
	# Y	playlist 		resume 						<playlist>	
	# Y	playlist 		save 						<playlist>	
	# Y	playlist		loadalbum|playalbum			<genre>						<artist>	<album>		<songtitle>
	# Y	playlist		addalbum					<genre>						<artist>	<album>		<songtitle>
	# Y	playlist		insertalbum					<genre>						<artist>	<album>		<songtitle>
	# Y	playlist		deletealbum					<genre>						<artist>	<album>		<songtitle>
	# Y	playlist 		clear	
	# Y	playlist		zap							<index>
	# Y	playlist		name						?
	# Y	playlist		url							?
	# Y	playlist		modified					?
	# Y	playlist		index|jump 					<index|?>	
	# Y	playlist		genre						<index>						?
	# Y	playlist		artist						<index>						?
	# Y	playlist		album						<index>						?
	# Y	playlist		title						<index>						?
	# Y	playlist		duration					<index>						?
	# Y	playlist		tracks						?
	# Y	playlist		shuffle						<0|1|2|?|>
	# Y	playlist		repeat						<0|1|2|?|>


	
	if (!defined($p0)) {
		# ignore empty commands
	
# The commands below do not require a client

	} elsif ($p0 eq "player") {
		if (!defined($p1)) {
		
		} elsif ($p1 eq "count") {
			$p2 = Slim::Player::Client::clientCount();
		} elsif ($p1 eq "name" || $p1 eq "address" || $p1 eq "ip" || $p1 eq "id" || $p1 eq "model" || $p1 eq "displaytype") {
			my $p2client;
			
			# were we passed an ID?
			if (defined $p2 && Slim::Player::Client::getClient($p2)) {
				$p2client = Slim::Player::Client::getClient($p2);
			} else {
			
			# otherwise, try for an index
				my @clients = Slim::Player::Client::clients();
				if (defined $p2 && defined $clients[$p2]) {
					$p2client = $clients[$p2];
				}
			}
			
			if (defined $p2client) {
				if ($p1 eq "name") {
					$p3 = $p2client->name();
				} elsif ($p1 eq "address" || $p1 eq "id") {
					$p3 = $p2client->id();
				} elsif ($p1 eq "ip") {
					$p3 = $p2client->ipport();
				} elsif ($p1 eq "model") {
					$p3 = $p2client->model();
				} elsif ($p1 eq "displaytype") {
					$p3 = $p2client->vfdmodel();
				}
			}
		}
		$client = undef;
		
	} elsif ($p0 eq "pref") {
		if (defined($p2) && $p2 ne '?' && !$::nosetup) {
			Slim::Utils::Prefs::set($p1, $p2);
		}
		$p2 = Slim::Utils::Prefs::get($p1);
		$client = undef;
		
	} elsif ($p0 eq "rescan") {
		Slim::Music::Import::startScan();
		$client = undef;
		
	} elsif ($p0 eq "wipecache") {
		Slim::Music::Info::wipeDBCache();
		Slim::Music::Import::startScan();
		$client = undef;
		
	} elsif ($p0 eq "info") {
		if (!defined($p1)) {
		} elsif ($p1 eq "total") {
			if ($p2 eq "genres") {
			   $p3 = Slim::Music::Info::genreCount([],[],[],[]);
			} elsif ($p2 eq "artists") {
			   $p3 = Slim::Music::Info::artistCount([],[],[],[]);
			} elsif ($p2 eq "albums") {
			   $p3 = Slim::Music::Info::albumCount([],[],[],[]);
			} elsif ($p2 eq "songs") {
			   $p3 = Slim::Music::Info::songCount([],[],[],[]);
			}
		}
		$client = undef;

	} elsif ($p0 eq "debug") {
		if ($p1 =~ /^d_/) {
			my $debugsymbol = "::" . $p1;
			no strict 'refs';
			if (!defined($p2)) {
				$$debugsymbol = ! $$debugsymbol;
				$p2 = $$debugsymbol;

			} elsif ($p2 eq "?")  {
				$p2 = $$debugsymbol;
			} else {
				$$debugsymbol = $p2;
			}	
			use strict 'refs';
			if (!$p2) { $p2 = 0; }
		}
		$client = undef;
		
	} elsif ($p0 eq "players") {

		$pushParams = 0;
	
		push @returnArray, $p0;
		if (defined($p1)) {
			push @returnArray, $p1;
		}
		if (defined($p2)) {
			push @returnArray, $p2;
		}
		
		my $count = Slim::Player::Client::clientCount();
		push @returnArray, "count:" . $count;
		
		my ($valid, $start, $end) = normalize(scalar($p1), scalar($p2), $count);
	
		if($valid) {
			my $idx = $start;
			my @players = Slim::Player::Client::clients();
	
			if (scalar(@players) > 0) {
				foreach my $eachclient (@players[$start..$end]) {
					push @returnArray, "playerindex:" . $idx;
					push @returnArray, "playerid:" . $eachclient->id();
					push @returnArray, "ip:" . $eachclient->ipport();
					push @returnArray, "name:" . $eachclient->name();
					push @returnArray, "model:" . $eachclient->model(); #squeezebox, softsqueeze, slimp3
					push @returnArray, "displaytype:" . $eachclient->vfdmodel(); 
					push @returnArray, "connected:" . ($eachclient->connected() || 0);
					$idx++;
				}	
			}
		}
		$client = undef;
		
	} elsif ($p0 eq "genres") {
		$pushParams = 0;
	
		my $search = "*";

		my %params = parseparams($parrayref, \@returnArray);
		
		$search = $params{'search'} if defined($params{'search'});
		
		if (Slim::Utils::Misc::stillScanning()) {
			push @returnArray, "rescan:1";
		}

		my @items = Slim::Music::Info::genres(Slim::Web::Pages::searchStringSplit($search), [], [], []);

		push @returnArray, "count:" . scalar(@items);

		my ($valid, $start, $end) = normalize(scalar($p1), scalar($p2), scalar(@items));

		if($valid) {
			foreach my $eachitem (@items[$start..$end]) {
				push @returnArray, "genre:" . $eachitem;
			}
		}
		$client = undef;
		
	} elsif ($p0 eq "artists" || $p0 eq "albums") {
		$pushParams = 0;
	
		my $search = "*";
		my $genre = "*";
		my $artist = "*";
		my $album = "*";

		my %params = parseparams($parrayref, \@returnArray);
		
		$search = $params{'search'} if defined($params{'search'});
		$genre = $params{'genre'} if defined ($params{'genre'});
		$artist = $params{'artist'} if defined($params{'artist'});
		$album = $params{'album'} if defined($params{'album'});


		if (Slim::Utils::Misc::stillScanning()) {
			push @returnArray, "rescan:1";
		}


		my @items;
		my $label;
		
		if($p0 eq "artists") {
			$label = "artist:";
			@items = Slim::Music::Info::artists([$genre],Slim::Web::Pages::searchStringSplit($search), [$album], []);
		} else {
			$label = "album:";
			@items = Slim::Music::Info::albums([$genre], [$artist], Slim::Web::Pages::searchStringSplit($search), []);
		}

		push @returnArray, "count:" . scalar(@items);

		my ($valid, $start, $end) = normalize(scalar($p1), scalar($p2), scalar(@items));

		if($valid) {
			foreach my $eachitem (@items[$start..$end]) {
				push @returnArray, $label . $eachitem;
			}
		}
		$client = undef;
		
	} elsif ($p0 eq "titles" || $p0 eq "songs") {
		$pushParams = 0;
	
		my $search = "*";
		my $genre = "*";
		my $album = "*";
		my $artist = "*";
		my $sort = "titles";
		my $tags = "galdp";
		
		my %params = parseparams($parrayref, \@returnArray);
		
		$search = $params{'search'} if defined($params{'search'});
		$genre = $params{'genre'} if defined ($params{'genre'});
		$artist = $params{'artist'} if defined($params{'artist'});
		$album = $params{'album'} if defined($params{'album'});
		$sort = $params{'sort'} if defined($params{'sort'});
		$tags = $params{'tags'} if defined($params{'tags'});
		
		if($sort eq "tracks") {
			$tags = $tags . "t";
		}
		
		if (Slim::Utils::Misc::stillScanning()) {
			push @returnArray, "rescan:1";
		}

		my @items = Slim::Music::Info::songs([$genre], [$artist], [$album], Slim::Web::Pages::searchStringSplit($search), ($sort eq "titles"));
		
		push @returnArray, "count:" . scalar(@items);

		my ($valid, $start, $end) = normalize(scalar($p1), scalar($p2), scalar(@items));
		my $cur = $start;

		if($valid) {
			foreach my $eachitem (@items[$start..$end]) {
				push @returnArray, "index:" . $cur;
				$cur++;
				push @returnArray, pushSong($eachitem, $tags);
				::idleStreams();
			}
		}
		$client = undef;
		
	} elsif ($p0 eq "playlists") {
		$pushParams = 0;
#		${$pcmdIsQuery} = 1;
	
		my $tags = "galdp";
		my $dir = "__playlists";
		my $search = "*";
		
		my %params = parseparams($parrayref, \@returnArray);
		
		$search = $params{'search'} if defined($params{'search'});
		$tags = $params{'tags'} if defined($params{'tags'});
		$dir = $params{'dir'} if defined($params{'dir'});
		
		my $fulldir = Slim::Utils::Misc::virtualToAbsolute($dir);
		my $badconfig = 0;
		my $dirscan = 1;
		
		if (!Slim::Utils::Prefs::get("playlistdir") && !(Slim::Music::iTunes::useiTunesLibrary() || Slim::Music::MoodLogic::useMoodLogic())) {
#			$::d_http && msg("no valid playlists directory!!\n");
			$badconfig = 1;
		}
		
		if (!$fulldir || !Slim::Music::Info::isList($fulldir)) {	
			# check if we're just showing itunes playlists
			if (Slim::Music::iTunes::useiTunesLibrary() || Slim::Music::MoodLogic::useMoodLogic()) {
				$dirscan = 0;
			} else {
#				$::d_http && msg("the selected playlist $fulldir isn't good!!.\n");
				$badconfig = 1;
			}
		}

		if (Slim::Utils::Misc::stillScanning()) {
			push @returnArray, "rescan:1";
		}
		
		if ($badconfig)
		{
			push @returnArray, "count:0";
		}
		else
		{
			# Get all playlist items
			my  $items = [];
					
			if($dirscan) {
				Slim::Utils::Scan::addToList($items, $fulldir, 0, undef, undef);
			}

			if ($dir eq '__playlists' && (Slim::Music::MoodLogic::useMoodLogic() || Slim::Music::iTunes::useiTunesLibrary())) {
				push @$items, @{Slim::Music::Info::playlists()};
			}
			
			# Perform search if necessary
			if(scalar @{$items} && $search ne "*")
			{
				my $search_pats = Slim::Music::Info::filterPats(Slim::Web::Pages::searchStringSplit($search));
				
				my @found;
				ITEM: foreach my $eachitem (@{$items}) {
					my $text = Slim::Music::Info::standardTitle(undef, $eachitem);
					foreach my $pat (@{$search_pats}) {
						if ($text !~ $pat) {
							::idleStreams();
							next ITEM;
						}
					}
					push @found, $eachitem;
				}
				
				my $foundqty = scalar(@found);
				@$items = @found;
			}
			
			
			my $numitems = scalar @{$items};
			push @returnArray, "count:" . $numitems;
			
			my ($valid, $start, $end) = normalize(scalar($p1), scalar($p2), $numitems);
			my $cur = $start ;

			if($valid) {
				foreach my $eachitem (@{$items}[$start..$end]) {
					::idleStreams();
					
					next if $eachitem =~ /__current.m3u$/;
					
					push @returnArray, "index:" . ($cur);
					
					
					if (Slim::Music::Info::isList($eachitem)) {
						#is a dir
						push @returnArray, "item:" . Slim::Music::Info::standardTitle(undef, $eachitem);
						push @returnArray, "dir:" . Slim::Utils::Misc::descendVirtual($dir, $eachitem, $cur);
					} elsif (Slim::Music::Info::isSong($eachitem)) {
						#is a song
						push @returnArray, pushSong($eachitem, $tags);
					}
	
					$cur++;
				}
			}
		}		
		$client = undef;

	} elsif ($p0 eq "songinfo") {
		$pushParams = 0;
	
		my $path;
		my $tags = "abcdefghijklmnoqrstuvwxyz"; #all letter EXCEPT p
		
		my %params = parseparams($parrayref, \@returnArray);
		
		$path = $params{'path'} if defined($params{'path'});
		$tags = $params{'tags'} if defined($params{'tags'});
				
		if (Slim::Utils::Misc::stillScanning()) {
			push @returnArray, "rescan:1";
		}
		
		if(Slim::Music::Info::isSong($path)) {

			my @items = pushSong($path, $tags);
			
			push @returnArray, "count:" . scalar(@items);
	
			my ($valid, $start, $end) = normalize(scalar($p1), scalar($p2), scalar(@items));
	
			if($valid) {
				foreach my $eachitem (@items[$start..$end]) {
					push @returnArray, $eachitem;
				}
			}
		} else {
			push @returnArray, "count:0";
		}
		$client = undef;
		
		
		
	#the following commands require a valid client to be specified
	} elsif ($client) {

		if ($p0 eq "playerpref") {
			if (defined($p2) && $p2 ne '?' && !$::nosetup) {
				Slim::Utils::Prefs::clientSet($client, $p1, $p2);
			}
			$p2 = Slim::Utils::Prefs::clientGet($client, $p1);

		} elsif ($p0 eq "play") {
			Slim::Player::Source::playmode($client, "play");
			Slim::Player::Source::rate($client,1);
			$client->update();

		} elsif ($p0 eq "pause") {
			if (defined($p1)) {
				if ($p1 && $client->playmode eq "play") {
					Slim::Player::Source::playmode($client, "pause");
				} elsif (!$p1 && $client->playmode eq "pause") {
					Slim::Player::Source::playmode($client, "resume");
				} elsif (!$p1 && $client->playmode eq "stop") {
					Slim::Player::Source::playmode($client, "play");
				}
			} else {
				if ($client->playmode eq "pause") {
					Slim::Player::Source::playmode($client, "resume");
				} elsif ($client->playmode eq "play") {
					Slim::Player::Source::playmode($client, "pause");
				} elsif ($client->playmode eq "stop") {
					Slim::Player::Source::playmode($client, "play");
				}
			}
			$client->update();

		} elsif ($p0 eq "rate") {
			if ($client->audioFilehandleIsSocket) {
				Slim::Player::Source::rate($client, 1);
			} elsif (!defined($p1) || $p1 eq "?") {
				$p1 = Slim::Player::Source::rate($client);
			} else {
				Slim::Player::Source::rate($client,$p1);
			}

		} elsif ($p0 eq "stop") {
			Slim::Player::Source::playmode($client, "stop");
			$client->update();

		} elsif ($p0 eq "mode") {
			if (!defined($p1) || $p1 eq "?") {
				$p1 = Slim::Player::Source::playmode($client);
			} else {
				if ($client->playmode eq "pause" && $p1 eq "play") {
					Slim::Player::Source::playmode($client, "resume");
				} else {
					Slim::Player::Source::playmode($client, $p1);
				}
				$client->update();
			}

		} elsif ($p0 eq "sleep") {
			if ($p1 eq "?") {
				$p1 = $client->sleepTime() - Time::HiRes::time();
				if ($p1 < 0) {
					$p1 = 0;
				}
			} else {
				Slim::Utils::Timers::killTimers($client, \&gotosleep);
				
				if ($p1 != 0) {
					Slim::Utils::Timers::setTimer($client, Time::HiRes::time() + $p1, \&gotosleep);
					$client->sleepTime(Time::HiRes::time() + $p1);
					$client->currentSleepTime($p1 / 60);
				} else {
					$client->sleepTime(0);
					$client->currentSleepTime(0);
				}
			}

		} elsif ($p0 eq "gototime" || $p0 eq "time") {
			if ($p1 eq "?") {
				$p1 = Slim::Player::Source::songTime($client);
			} else {
				Slim::Player::Source::gototime($client, $p1);
			}

		} elsif ($p0 eq "duration") {
			$p1 = Slim::Music::Info::durationSeconds(Slim::Player::Playlist::song($client)) || 0;

		} elsif ($p0 eq "artist") {
			$p1 = Slim::Music::Info::artist(Slim::Player::Playlist::song($client)) || 0;

		} elsif ($p0 eq "album") {
			$p1 = Slim::Music::Info::album(Slim::Player::Playlist::song($client)) || 0;

		} elsif ($p0 eq "title") {
			$p1 = Slim::Music::Info::title(Slim::Player::Playlist::song($client)) || 0;

		} elsif ($p0 eq "genre") {
			$p1 = Slim::Music::Info::genre(Slim::Player::Playlist::song($client)) || string('NO_GENRE');

		} elsif ($p0 eq "path") {
			$p1 = Slim::Player::Playlist::song($client) || 0;

		} elsif ($p0 eq "connected") {
			$p1 = $client->connected() || 0;

		} elsif ($p0 eq "signalstrength") {
			$p1 = $client->signalStrength() || 0;

		} elsif ($p0 eq "power") {
			if (!defined $p1) {
				if (Slim::Player::Sync::isSynced($client)) {
					syncFunction($client, !$client->power(), "power", undef);
				}
				$client->power(!$client->power());
			} elsif ($p1 eq "?") {
				$p1 = $client->power();
			} else {
				if (Slim::Player::Sync::isSynced($client)) {
					syncFunction($client, $p1, "power", undef);
				}
				$client->power($p1);
			}
			
		} elsif ($p0 eq "sync") {
			if (!defined $p1){
			} elsif ($p1 eq "?") {
				$p1 = Slim::Player::Sync::syncIDs($client);
			} elsif ($p1 eq "-") {
				Slim::Player::Sync::unsync($client);
			} else {
				my $buddy;
				if (Slim::Player::Client::getClient($p1)) {
					$buddy = Slim::Player::Client::getClient($p1);
				} else {
					my @clients = Slim::Player::Client::clients();
					if (defined $clients[$p1]) {
						$buddy = $clients[$p1];
					}
				}
				Slim::Player::Sync::sync($buddy, $client) if defined $buddy;
			}
			
		} elsif ($p0 eq "playlist") {
			# here are all the commands that add/insert/replace songs/directories/playlists on the current playlist
			if ($p1 =~ /^(play|load|append|add|resume|insert|insertlist)$/) {
			
				my $jumptoindex = undef;
				my $path = $p2;
				if ($path) {
					if (!-e $path && !(Slim::Music::Info::isITunesPlaylistURL($path) || Slim::Music::Info::isMoodLogicPlaylistURL($path))) {
						my $easypath = catfile(Slim::Utils::Prefs::get('playlistdir'), basename ($p2) . ".m3u");
						if (-e $easypath) {
							$path = $easypath;
						} else {
							$easypath = catfile(Slim::Utils::Prefs::get('playlistdir'), basename ($p2) . ".pls");
							if (-e $easypath) {
								$path = $easypath;
							}
						}
					}
					
					if ($p1 =~ /^(play|load|resume)$/) {
						Slim::Player::Source::playmode($client, "stop");
						Slim::Player::Playlist::clear($client);
						my $fixpath = Slim::Utils::Misc::fixPath($path);
						$client->currentPlaylist($fixpath);
						Slim::Music::Info::setTitle($fixpath, $p3) if defined $p3;
						$client->currentPlaylistModified(0);
					} else {
						$client->currentPlaylistModified(1);
					}
					
					$path = Slim::Utils::Misc::virtualToAbsolute($path);
					
					if ($p1 =~ /^(play|load)$/) { 
						$jumptoindex = 0;
					} elsif ($p1 eq "resume" && Slim::Music::Info::isM3U($path)) {
						# do nothing to the index if we can't open the list
						my $playlist = new FileHandle($path, "r");
						if ($playlist) {
							# retrieve comment with track number in it
							$jumptoindex = $playlist->getline;
							if ($jumptoindex =~ /^#CURTRACK (\d+)$/) {
								$jumptoindex = $1;
							} else {
								$jumptoindex = 0;
							}
							close $playlist;
						}
					}
					
					if ($p1 =~ /^(insert|insertlist)$/) {
						my $playListSize = Slim::Player::Playlist::count($client);
						my @dirItems;
						Slim::Utils::Scan::addToList(\@dirItems, $path, 1, undef);
						Slim::Utils::Scan::addToList(
							Slim::Player::Playlist::playList($client)
							, $path
							, 1
							, undef
							, \&insert_done
							, $client
							, $playListSize
							, scalar(@dirItems)
							,$callbackf
							, $callbackargs
						);
					} else {
						Slim::Utils::Scan::addToList(
							Slim::Player::Playlist::playList($client)
							, $path
							, 1
							, undef
							, \&load_done
							, $client
							, $jumptoindex
							, $callbackf
							, $callbackargs
							);
					}
					
					$callcallback = 0;
					$p2 = $path;
				}
			
			} elsif ($p1 eq "loadalbum" | $p1 eq "playalbum") {
				Slim::Player::Source::playmode($client, "stop");
				Slim::Player::Playlist::clear($client);
				push(@{Slim::Player::Playlist::playList($client)}, Slim::Music::Info::songs(singletonRef($p2), singletonRef($p3), singletonRef($p4), singletonRef($p5), $p6));
				Slim::Player::Playlist::reshuffle($client);
				Slim::Player::Source::jumpto($client, 0);
				$client->currentPlaylist(undef);
			
			} elsif ($p1 eq "addalbum") {
				push(@{Slim::Player::Playlist::playList($client)}, Slim::Music::Info::songs(singletonRef($p2), singletonRef($p3), singletonRef($p4), singletonRef($p5), $p6));
				Slim::Player::Playlist::reshuffle($client);
				$client->currentPlaylistModified(1);
			
			} elsif ($p1 eq "insertalbum") {
				my @songs = Slim::Music::Info::songs(singletonRef($p2), singletonRef($p3), singletonRef($p4), singletonRef($p5), $p6);
				my $playListSize = Slim::Player::Playlist::count($client);
				my $size = scalar(@songs);
				push(@{Slim::Player::Playlist::playList($client)}, @songs);
				insert_done($client, $playListSize, $size);
				#Slim::Player::Playlist::reshuffle($client);
				$client->currentPlaylistModified(1);
			
			} elsif ($p1 eq "save") {
				if ( Slim::Utils::Prefs::get('playlistdir')) {
					# just use the filename to avoid nasties
					my $savename = basename ($p2);
					# save the current playlist position as a comment at the head of the list
					my $annotatedlistRef;
					if (Slim::Utils::Prefs::get('saveShuffled')) {
						foreach my $shuffleitem (@{Slim::Player::Playlist::shuffleList($client)}) {
							push (@$annotatedlistRef,@{Slim::Player::Playlist::playList($client)}[$shuffleitem]);
						}
					} else {
						$annotatedlistRef = Slim::Player::Playlist::playList($client);
					}
					Slim::Formats::Parse::writeM3U( 
						$annotatedlistRef 
						,undef,
						,catfile(Slim::Utils::Prefs::get('playlistdir') 
						,$savename . ".m3u") 
						,1 
						,Slim::Player::Source::currentSongIndex($client)
					);
				}
			
			} elsif ($p1 eq "deletealbum") {
				my @listToRemove=Slim::Music::Info::songs(singletonRef($p2),singletonRef($p3),singletonRef($p4),singletonRef($p5),$p6);
				Slim::Player::Playlist::removeMultipleTracks($client,\@listToRemove);
				$client->currentPlaylistModified(1);
			
			} elsif ($p1 eq "deleteitem") {
				if (defined($p2) && $p2 ne '') {
					$p2 = Slim::Utils::Misc::virtualToAbsolute($p2);
					my $contents;
					if (!Slim::Music::Info::isList($p2)) {
						Slim::Player::Playlist::removeMultipleTracks($client,[$p2]);
					} elsif (Slim::Music::Info::isDir($p2)) {
						Slim::Utils::Scan::addToList(\@{$contents}, $p2, 1);
						Slim::Player::Playlist::removeMultipleTracks($client,\@{$contents});
					} else {
						$contents = Slim::Music::Info::cachedPlaylist($p2);
						if (defined($contents)) {
							$contents = [Slim::Music::Info::cachedPlaylist($p2)];
						} else {
							my $playlist_filehandle;
							if (!open($playlist_filehandle, Slim::Utils::Misc::pathFromFileURL($p2))) {
								($::d_files || $::d_command) && msg("Couldn't open playlist file $p2 : $!\n");
								$playlist_filehandle = undef;
							} else {
								$contents = [Slim::Formats::Parse::parseList($p2,$playlist_filehandle,dirname($p2))];
							}
						}
						if (defined($contents)) {
							Slim::Player::Playlist::removeMultipleTracks($client,$contents);
						}
					}
				$client->currentPlaylistModified(1);
				}
			
			} elsif ($p1 eq "repeat") {
				# change between repeat values 0 (don't repeat), 1 (repeat the current song), 2 (repeat all)
				if (!defined($p2)) {
					Slim::Player::Playlist::repeat($client, (Slim::Player::Playlist::repeat($client) + 1) % 3);
				} elsif ($p2 eq "?") {
					$p2 = Slim::Player::Playlist::repeat($client);
				} else {
					Slim::Player::Playlist::repeat($client, $p2);
				}
			
			} elsif ($p1 eq "shuffle") {
				if (!defined($p2)) {
					my $nextmode=(1,2,0)[Slim::Player::Playlist::shuffle($client)];
					Slim::Player::Playlist::shuffle($client, $nextmode);
					Slim::Player::Playlist::reshuffle($client);
				} elsif ($p2 eq "?") {
					$p2 = Slim::Player::Playlist::shuffle($client);
				} else {
					Slim::Player::Playlist::shuffle($client, $p2);
					Slim::Player::Playlist::reshuffle($client);
				}
			
			} elsif ($p1 eq "clear") {
				Slim::Player::Playlist::clear($client);
				Slim::Player::Source::playmode($client, "stop");
				$client->currentPlaylist(undef);
			
			} elsif ($p1 eq "move") {
				Slim::Player::Playlist::moveSong($client, $p2, $p3);
				$client->currentPlaylistModified(1);
			
			} elsif ($p1 eq "delete") {
				if (defined($p2)) {
					Slim::Player::Playlist::removeTrack($client,$p2);
				}
				$client->currentPlaylistModified(1);
			
			} elsif (($p1 eq "jump") || ($p1 eq "index")) {
				if ($p2 eq "?") {
					$p2 = Slim::Player::Source::currentSongIndex($client);
				} else {
						Slim::Player::Source::jumpto($client, $p2);
				}
				
			} elsif ($p1 eq "name") {
				$p2 = Slim::Music::Info::standardTitle($client,$client->currentPlaylist());
				
			} elsif ($p1 eq "url") {
				$p2 = $client->currentPlaylist();

			} elsif ($p1 eq "modified") {
				$p2 = $client->currentPlaylistModified();
				
			} elsif ($p1 eq "tracks") {
				$p2 = Slim::Player::Playlist::count($client);
				
			} elsif ($p1 eq "duration") {
				$p3 = Slim::Music::Info::durationSeconds(Slim::Player::Playlist::song($client,$p2)) || 0;
				
			} elsif ($p1 eq "artist") {
				$p3 = Slim::Music::Info::artist(Slim::Player::Playlist::song($client,$p2)) || 0;
				
			} elsif ($p1 eq "album") {
				$p3 = Slim::Music::Info::album(Slim::Player::Playlist::song($client,$p2)) || 0;
				
			} elsif ($p1 eq "title") {
				$p3 = Slim::Music::Info::title(Slim::Player::Playlist::song($client,$p2)) || 0;
				
			} elsif ($p1 eq "genre") {
				$p3 = Slim::Music::Info::genre(Slim::Player::Playlist::song($client,$p2)) || string('NO_GENRE');
				
			} elsif ($p1 eq "path") {
				$p3 = Slim::Player::Playlist::song($client,$p2) || 0;
			
			} elsif ($p1 eq "zap") {
				my $zapped=catfile(Slim::Utils::Prefs::get('playlistdir'), string('ZAPPED_SONGS') . '.m3u');
				my $zapsong = Slim::Player::Playlist::song($client,$p2);
				my $zapindex = $p2 || Slim::Player::Source::currentSongIndex($client);;
				#Remove from current playlist
				if (Slim::Player::Playlist::count($client) > 0) {
					Slim::Control::Command::execute($client, ["playlist", "delete", $zapindex]);
				}
				# Append the zapped song to the zapped playlist
				# This isn't as nice as it should be, but less work than loading and rewriting the whole list
				my $zapref = new FileHandle $zapped, "a";
				if ($zapref) {
					my @zaplist = ($zapsong);
					my $zapitem = Slim::Formats::Parse::writeM3U(\@zaplist);
					print $zapref $zapitem;
					close $zapref;
				} else {
					msg("Could not open $zapped for writing.\n");
				}
				$client->currentPlaylistModified(1);
			}

		} elsif ($p0 eq "mixer") {
			if ($p1 eq "volume") {
				my $newvol;
				my $oldvol = $client->volume();
				if ($p2 eq "?") {
					$p2 = $oldvol;
				} else {
					if($oldvol < 0) {
						# volume was previously muted
						$oldvol *= -1;      # un-mute volume
					} 
					
					if ($p2 =~ /^[\+\-]/) {
						$newvol = $oldvol + $p2;
					} else {
						$newvol = $p2;
					}
					
					$newvol = $client->volume($newvol);
					
					if (Slim::Player::Sync::isSynced($client)) {syncFunction($client, $newvol, "volume",\&setVolume);};
				}
				
			} elsif ($p1 eq "muting") {
				my $vol = $client->volume();
				my $fade;
				
				if($vol < 0) {
					# need to un-mute volume
					$::d_command && msg("Unmuting, volume is $vol.\n");
					Slim::Utils::Prefs::clientSet($client, "mute", 0);
					$fade = 0.3125;
				} else {
					# need to mute volume
					$::d_command && msg("Muting, volume is $vol.\n");
					Slim::Utils::Prefs::clientSet($client, "mute", 1);
					$fade = -0.3125;
				}
				# work around to bug 743, use mute() instead of fade_volume.
				# see also syncFunction...
				
#				$client->fade_volume($fade, \&mute, [$client]);
				$client->mute();
				if (Slim::Player::Sync::isSynced($client)) {syncFunction($client, $fade, "mute", undef);};
				
			} elsif ($p1 eq "balance") {
				# unsupported yet
				
			} elsif ($p1 eq "treble") {
				my $newtreb;
				my $oldtreb = $client->treble();
				if ($p2 eq "?") {
					$p2 = $oldtreb;
				} else {
				
					if ($p2 =~ /^[\+\-]/) {
						$newtreb = $oldtreb + $p2;
					} else {
						$newtreb = $p2;
					}
					$newtreb = $client->treble($newtreb);
					if (Slim::Player::Sync::isSynced($client)) {syncFunction($client, $newtreb, "treble",\&setTreble);};
				}
				
			} elsif ($p1 eq "bass") {
				my $newbass;
				my $oldbass = $client->bass();
				if ($p2 eq "?") {
					$p2 = $oldbass;
				} else {
#				
					if ($p2 =~ /^[\+\-]/) {
						$newbass = $oldbass + $p2;
					} else {
						$newbass = $p2;
					}
					$newbass = $client->bass($newbass);
					if (Slim::Player::Sync::isSynced($client)) {syncFunction($client, $newbass, "bass",\&setBass);};
				}
				
			} elsif ($p1 eq "pitch") {
				my $newpitch;
				my $oldpitch = $client->pitch();
				if ($p2 eq "?") {
					$p2 = $oldpitch;
				} else {
					if ($p2 =~ /^[\+\-]/) {
						$newpitch = $oldpitch + $p2;
					} else {
						$newpitch = $p2;
					}
					$newpitch = $client->pitch($newpitch);
					if (Slim::Player::Sync::isSynced($client)) {syncFunction($client, $newpitch, "pitch",\&setPitch);};
				}
			}
			
		} elsif ($p0 eq "displaynow") {
			if ($p1 eq "?" && $p2 eq "?") {
				$p1 = $client->prevline1();
				$p2 = $client->prevline2();
			} 
			
		} elsif ($p0 eq "linesperscreen") {
				$p1 = $client->linesPerScreen();
			
		} elsif ($p0 eq "display") {
			if ($p1 eq "?" && $p2 eq "?") {
				my ($line1, $line2) = Slim::Display::Display::curLines($client);
				$p1 = $line1;
				$p2 = $line2;
			} else {
				Slim::Buttons::ScreenSaver::wakeup($client);
				$client->showBriefly($p1, $p2, $p3, $p4);
			}
			
		} elsif ($p0 eq "button") {
			# all buttons now go through execute()
			Slim::Hardware::IR::executeButton($client, $p1, $p2, undef, defined($p3) ? $p3 : 1);
			
		} elsif ($p0 eq "ir") {
			# all ir signals go through execute()
			Slim::Hardware::IR::processIR($client, $p1, $p2);
			
	#Extended CLI API STATUS		
		} elsif ($p0 eq "status") {
			$pushParams = 0;

			my $tags = "galdp";
			
			my %params = parseparams($parrayref, \@returnArray);
			
			$tags = $params{'tags'} if defined($params{'tags'});
			
			my $connected = $client->connected() || 0;
			my $power = $client->power();
			my $repeat=Slim::Player::Playlist::repeat($client);
			my $shuffle=Slim::Player::Playlist::shuffle($client);
	    	my $songcount = Slim::Player::Playlist::count($client);
	    	my $idx = 0;
		    	
			if (Slim::Utils::Misc::stillScanning()) {
				push @returnArray, "rescan:1";
			}
			
			push @returnArray, "player_name:" . $client->name();
			push @returnArray, "player_connected:" . $connected;
			
			push @returnArray, "power:".$power;
			
			if($client->model() eq "squeezebox") {
				push @returnArray, "signalstrength:".($client->signalStrength() || 0);
			}
			
			if ($power) {
			
				#push @returnArray, "player_mode:".Slim::Buttons::Common::mode($client);
			
		    	push @returnArray, "mode:" . Slim::Player::Source::playmode($client);
				if (Slim::Player::Playlist::song($client)) { 
					my $dur = Slim::Music::Info::durationSeconds(Slim::Player::Playlist::song($client));
					if ($dur) {
				    	push @returnArray, "rate:".Slim::Player::Source::rate($client); #(>1 ffwd, <0 rew else norm)
		    			push @returnArray, "time:".Slim::Player::Source::songTime($client);
		    			push @returnArray, "duration:".$dur;
					}
				}
				
				if ($client->currentSleepTime()) {
					my $sleep = $client->sleepTime() - Time::HiRes::time();
		    		push @returnArray, "sleep:" . $client->currentSleepTime()*60;
		    		push @returnArray, "will_sleep_in:" . ($sleep < 0 ? 0 : $sleep);
				}
				
				if (Slim::Player::Sync::isSynced($client)) {
					my $master = Slim::Player::Sync::masterOrSelf($client);
						push @returnArray, "sync_master:" . $master->id();
					my @slaves = Slim::Player::Sync::slaves($master);
					my @sync_slaves = map{$_->id()} @slaves;
					push @returnArray, "sync_slaves:" . join(" ", @sync_slaves);
				}
			
		    	push @returnArray, "mixer volume:".$client->volume();
		    	push @returnArray, "mixer treble:".$client->treble();
		    	push @returnArray, "mixer bass:".$client->bass();
		    	if($client->model() ne "slimp3") {
		    		push @returnArray, "mixer pitch:".$client->pitch();
		    	}
		    	#push @returnArray, "mixer balance:";


		    	push @returnArray, "playlist repeat:".$repeat; #(0 no, 1 title, 2 all)
		    	push @returnArray, "playlist shuffle:".$shuffle; #(0 no, 1 title, 2 albums)
		    
				
				if($songcount>0){
					$idx = Slim::Player::Source::currentSongIndex($client);
					push @returnArray, "playlist_cur_index:".($idx);
				}
		    	push @returnArray, "playlist_tracks:".$songcount;
			}
			

			if ($songcount>0 && $power) {
			
				#we can return playlist data.
				#which mode are we in?
				my $modecurrent = 0;
				if(defined($p1) && ($p1 eq "-")) {
					$modecurrent = 1;
				}
				
				
				#if repeat is 1 (song) and modecurrent, then show the current song
				if($modecurrent && ($repeat == 1)) {
					push @returnArray, "playlist index:".($idx);
					push @returnArray, pushSong(Slim::Player::Playlist::song($client, $idx), $tags);	
				} else {
					my $valid;
					my $start;
					my $end;
					
					if ($modecurrent) {
						($valid, $start, $end) = normalize(($idx), scalar($p2), $songcount);
					} else {
						($valid, $start, $end) = normalize(scalar($p1), scalar($p2), $songcount);
					}
		
					if($valid) {
					
						my $count = 0;
	
						for ($idx=$start; $idx<=$end; $idx++){
							$count++;
							push @returnArray, "playlist index:".($idx);
							push @returnArray, pushSong(Slim::Player::Playlist::song($client, $idx), $tags);
							::idleStreams() ;
						}
						
						if(($repeat == 2) && ($shuffle == 0) && ($count < scalar($p2))) {
							#wrap around the playlist...
							($valid, $start, $end) = normalize(0, (scalar($p2) - $count), $songcount);		
							if($valid) {
								for ($idx=$start; $idx<=$end; $idx++){
									push @returnArray, "playlist index:".($idx);
									push @returnArray, pushSong(Slim::Player::Playlist::song($client, $idx), $tags);
									::idleStreams() ;
								}
							}						
						}
					}
				}
			}
		} elsif ($p0 eq "listen") {
			$client = undef;
		} elsif ($p0 eq "exit") {
			$client = undef;
		} else {
			#treat anything we don't know as a query so that we don't bcast it!
#			${$pcmdIsQuery} = 1;
		}
	}				
		
	#extended CLI API calls do push their return values directly
	if ($pushParams) {
		if (defined($p0)) { push @returnArray, $p0 };
		if (defined($p1)) { push @returnArray, $p1 };
		if (defined($p2)) { push @returnArray, $p2 };
		if (defined($p3)) { push @returnArray, $p3 };
		if (defined($p4)) { push @returnArray, $p4 };
		if (defined($p5)) { push @returnArray, $p5 };
		if (defined($p6)) { push @returnArray, $p6 };
		if (defined($p7)) { push @returnArray, $p7 };
	}
	
	$callcallback && $callbackf && (&$callbackf(@$callbackargs, \@returnArray));

	executeCallback($client, \@returnArray);
	
	$::d_command && msg(" Returning array: $p0 (" .
			(defined $p1 ? $p1 : "") . ") (" .
			(defined $p2 ? $p2 : "") . ") (" .
			(defined $p3 ? $p3 : "") . ") (" .
			(defined $p4 ? $p4 : "") . ") (" .
			(defined $p5 ? $p5 : "") . ") (" .
			(defined $p6 ? $p6 : "") . ") (" .
			(defined $p7 ? $p7 : "") . ")\n");
	
	return @returnArray;
}

sub syncFunction {
	my $client = shift;
	my $newval = shift;
	my $setting = shift;
	my $controlRef = shift;
	
	my @buddies = Slim::Player::Sync::syncedWith($client);
	if (scalar(@buddies) > 0) {
		foreach my $eachclient (@buddies) {
			if (Slim::Utils::Prefs::clientGet($eachclient,'syncVolume')) {
				if ($setting eq "mute") {
					# work around to bug 743, use mute() instead of fade_volume
					# see also above 'mixer muting'
#					$eachclient->fade_volume($newval, \&mute, [$eachclient]);
					$eachclient->mute();
				} else {
					Slim::Utils::Prefs::clientSet($eachclient, $setting, $newval);
					&$controlRef($eachclient, $newval) if ($controlRef);
				}
				if ($setting eq "volume") {
					Slim::Display::Display::volumeDisplay($eachclient);
				}
			}
			if (Slim::Utils::Prefs::clientGet($client,'syncPower')) {
				if ($setting eq "power") {
					$eachclient->power($newval);
				}
			}
		}
	}
}

sub setVolume {
	my $client = shift;
	my $volume = shift;
	$client->volume($volume);
}

sub setBass {
	my $client = shift;
	my $bass = shift;
	$client->bass($bass);
}

sub setPitch {
	my $client = shift;
	my $pitch = shift;
	$client->pitch($pitch);
}

sub setTreble {
	my $client = shift;
	my $treble = shift;
	$client->treble($treble);
}

sub setExecuteCallback {
	my $callbackRef = shift;
	$executeCallbacks{$callbackRef} = $callbackRef;
}

sub clearExecuteCallback {
	my $callbackRef = shift;
	delete $executeCallbacks{$callbackRef};
}

sub executeCallback {
	my $client = shift;
	my $paramsRef = shift;

	no strict 'refs';
		
	foreach my $executecallback (keys %executeCallbacks) {
		$executecallback = $executeCallbacks{$executecallback};
		&$executecallback($client, $paramsRef);
	}
}

sub load_done {
	my ($client, $index, $callbackf, $callbackargs)=@_;
	# dont' keep current song on loading a playlist
	Slim::Player::Playlist::reshuffle($client,($client->playmode eq "play" || ($client->power && $client->playmode eq "pause"))?0:1);
	if (defined($index)) {
		Slim::Player::Source::jumpto($client, $index);
	}
	$callbackf && (&$callbackf(@$callbackargs));
	Slim::Control::Command::executeCallback($client, ['playlist','load_done']);
}

sub insert_done {
	my ($client, $listsize, $size, $callbackf, $callbackargs)=@_;
	my $i;
	my $playlistIndex = Slim::Player::Source::currentSongIndex($client)+1;
	my @reshuffled;

	if (Slim::Player::Playlist::shuffle($client)) {
		for ($i = 0; $i < $size; $i++) {
			push @reshuffled,$listsize+$i;
		};
		$client = Slim::Player::Sync::masterOrSelf($client);
		
		splice @{$client->shufflelist}, $playlistIndex, 0, @reshuffled;
	} else {
		Slim::Player::Playlist::moveSong($client, $listsize, $playlistIndex,$size);
		Slim::Player::Playlist::reshuffle($client);
	};
	Slim::Player::Playlist::refreshPlaylist($client);
	$callbackf && (&$callbackf(@$callbackargs));
	Slim::Control::Command::executeCallback($client, ['playlist','load_done']);
}

sub singletonRef {
	my $arg = shift;

	if (!defined($arg)) {
		return [];
	} elsif ($arg eq '*') {
		return [];
	} elsif ($arg) {
		return [$arg];
	} else {
		return [];
	}
}

sub gotosleep {
	my $client = shift;
	if ($client->isPlayer()) {
		$client->fade_volume(-60,\&turnitoff,[$client]);
	}
	$client->sleepTime(0);
	$client->currentSleepTime(0);
}

sub mute {
	my $client = shift;
	$client->mute();
}

sub turnitoff {
	my $client = shift;
	
	# Turn off quietly
	execute($client, ['stop', 0]);
	execute($client, ['power', 0]);
}


###################################################
#### Extended CLI API helper subs

sub pushSong {
	my $song = shift;
	my $tags = shift;
	
	my @returnArray = ();
	
	#a b c d E f g h i j k l m n o p q r S t u v X y z

	push @returnArray, "title:".Slim::Music::Info::title($song);
	
	if ($tags=~ /g/ && defined(Slim::Music::Info::genre($song))) {
		my $val = Slim::Music::Info::genre($song);
		if ($val eq ''){
			$val = string('NO_GENRE');
		}
		push @returnArray, "genre:".$val;
	}
	if ($tags=~ /a/ && defined(Slim::Music::Info::artist($song))) {
		push @returnArray, "artist:".Slim::Music::Info::artist($song);
	}
	if ($tags=~ /c/ && defined(Slim::Music::Info::composer($song))) {
		push @returnArray, "composer:".Slim::Music::Info::composer($song);
	}
	if ($tags=~ /b/ && defined(Slim::Music::Info::band($song))) {
		push @returnArray, "band:".Slim::Music::Info::band($song);
	}
	if ($tags=~ /u/ && defined(Slim::Music::Info::conductor($song))) {
		push @returnArray, "conductor:".Slim::Music::Info::conductor($song);
	}
	if ($tags=~ /l/ && defined(Slim::Music::Info::album($song))) {
		push @returnArray, "album:".Slim::Music::Info::album($song);
	}
	if ($tags=~ /d/) {
		push @returnArray, "duration:".Slim::Music::Info::durationSeconds($song);
	}
	if($tags=~ /i/ && defined(Slim::Music::Info::disc($song))){
		push @returnArray, "disc:".Slim::Music::Info::disc($song);
	}
	if($tags=~ /q/ && defined(Slim::Music::Info::discCount($song))){
		push @returnArray, "disccount:".Slim::Music::Info::discCount($song);
	}
	if($tags=~ /t/ && defined(Slim::Music::Info::trackNumber($song))) {
		push @returnArray, "track:".Slim::Music::Info::trackNumber($song);
	}
	if($tags=~ /y/ && defined(Slim::Music::Info::year($song))) {
		push @returnArray, "year:".Slim::Music::Info::year($song);
	}
	if($tags=~ /m/ && defined(Slim::Music::Info::bpm($song))) {
		push @returnArray, "bpm:".Slim::Music::Info::bpm($song);
	}
	if($tags=~ /k/ && defined(Slim::Music::Info::comment($song))) {
		push @returnArray, "comment:".Slim::Music::Info::comment($song);
	}
	if($tags=~ /o/ && defined(Slim::Music::Info::contentType($song))) {
		push @returnArray, "type:".string(uc(Slim::Music::Info::contentType($song)));
	}
	if($tags=~ /v/ && defined(Slim::Music::Info::tagVersion($song))) {
		push @returnArray, "tagversion:".Slim::Music::Info::tagVersion($song);
	}
	if($tags=~ /r/ && defined(Slim::Music::Info::bitrate($song))) {
		push @returnArray, "bitrate:".Slim::Music::Info::bitrate($song);
	}
	if($tags=~ /f/ && Slim::Music::Info::fileLength($song)) {
		push @returnArray, "filelength:".Slim::Music::Info::fileLength($song);
	}
	if($tags=~ /z/ && defined(Slim::Music::Info::digitalrights($song))) {
		push @returnArray, "drm:".Slim::Music::Info::digitalrights($song);
	}
	if ($tags=~ /j/) {
		my ($body, $type, $mtime) =  Slim::Music::Info::coverArt($song,'cover');
		if(defined($body)){
			push @returnArray, "coverart:1";
		}
	}
	if ($tags=~ /h/) {
		my ($body, $type, $mtime) =  Slim::Music::Info::coverArt($song,'thumb');
		if(defined($body)){
			push @returnArray, "coverthumb:1";
		}
	}
	if ($tags=~ /n/) {
		push @returnArray, "modtime:".Slim::Utils::Misc::longDateF(Slim::Music::Info::age($song)) . ", " . Slim::Utils::Misc::timeF(Slim::Music::Info::age($song));
	}	
	if($tags=~ /p/) {
		push @returnArray, "path:".$song;
	}
	
	return @returnArray;
}

sub normalize {
	my $from = shift;
	my $numofitems = shift;
	my $count = shift;
	
	my $start = 0;
	my $end = 0;
	my $valid = 0;
	
	if ($numofitems && $count) {
		my $lastidx = $count - 1;
		if($from > $lastidx) {
			return ($valid, $start, $end);
		}
		if($from < 0) {
			$from = 0;
		}
	
		$start = $from;
		$end = $start + $numofitems - 1;
	
		if ($end > $lastidx) {
			$end = $lastidx;
		}
		$valid = 1;
	}
	return ($valid, $start, $end);
}


sub parseparams {
	my $parrayref = shift;
	my $preturnarrayref = shift;
	
	my %params;
	
	foreach my $p (@$parrayref) {
		if ($p =~ /([^:]+):(.*)/) {
			if($1 eq "search") {
				$params{'search'} = $2;
			} elsif ($1 eq "genre") {
				$params{'genre'} = $2;
			} elsif ($1 eq "artist") {
				$params{'artist'} = $2;
			} elsif ($1 eq "album") {
				$params{'album'} = $2;
			} elsif ($1 eq "tags") {
				$params{'tags'} = $2;
			} elsif ($1 eq "sort") {
				$params{'sort'} = $2;
			} elsif ($1 eq "dir") {
				$params{'dir'} = $2;
			} elsif ($1 eq "path") {
				$params{'path'} = $2;
			}
		}
		push @$preturnarrayref, $p;
	}
	
	return %params;
}




1;
__END__

# Local Variables:
# tab-width:4
# indent-tabs-mode:t
# End:

